from inspect import isclass, ismethod, isfunction
from pandas import DataFrame

from eazydocs.generator.example import Example
from eazydocs.generator.generator import Generator
from eazydocs.generator.parameters import Parameters

from eazydocs.md.md_file import MDFile


def generate_docs(obj: object, skip_private: bool = True, append_to_file: bool = False, filename: str = None, filepath: str = None) -> str:
    if isclass(obj):
        docs = Generator(obj, skip_private).docs
    elif isfunction(obj) or ismethod(obj):
        docs = Parameters(obj, skip_private).params

    docs = docs.strip()

    if append_to_file != False:
        MDFile(filename, docs, filepath).append()

    return docs


def generate_example(
    df: DataFrame,
    df_shape: list[int] = [5, 5],
    code: str = "df",
    append_to_file: bool = False,
    filename: str = None,
    filepath: str = None,
    method_name: str = None,
) -> str:
    example = Example(code, df, df_shape).output

    if append_to_file != False:
        if filename == None:
            raise ValueError(
                "generate_example missing 1 required positional argument: 'filename'. Set 'append_to_file=False', if you would like the string output of this function."
            )

        if method_name != None:
            MDFile(filename, example, filepath).append_to_param(method_name)
        else:
            MDFile(filename, example, filepath)

    return example



from pathlib import Path


class MDFile:
    def __init__(self, filename: str = None, to_append: str = None, filepath: Path | str = None) -> None:
        if to_append == None:
            raise ValueError("Markdown() missing 1 required positional argument: 'to_append'")
        else:
            self.to_append = to_append

        self._set_path_attr(filepath)

        if filename != None:
            filename = self._check_filename(filename)
        else:
            filename = self._set_filename()

        self.filename = filename
        self.filepath = Path(self.cwd, filename)

    def append(self) -> None:
        to_append = f"\n\n{self.to_append}"
        with open(self.filepath, "+a") as f:
            f.write(to_append)

        print(f"Succesfully updated '{self.filename}' ({self.filepath})")

    def append_to_param(self, method_name: str) -> None:
        before, after = self._slice_contents_at_insert_position(method_name)
        to_append = self._generate_output_str(before, after)

        with open(self.filepath, "w") as f:
            f.write(to_append)

        print(f"Succesfully updated '{method_name}' in '{self.filename}' ({self.filepath})")

    def _set_path_attr(self, path: None | Path = None) -> None:
        if type(path) == str:
            path = Path(path)

        if path != None:
            self.p = path
            self.cwd = path
        else:
            p = Path()
            self.p = p
            self.cwd = p.cwd()

    def _check_filename(self, filename: str) -> str:
        filename = filename.strip()

        if filename[-3:] != ".md":
            filename += ".md"

        return filename

    def _set_filename(self) -> str:
        if Path(self.cwd, "README.md").exists():
            md_files = list(self.p.glob("README_*.md"))

            try:
                last_file = str(md_files.pop())
                last_filename = last_file.replace(".md", "").split("_")
                last_file_num = int(last_filename[-1])
                filename = f"README_{last_file_num+1}.md"
            except IndexError:
                filename = "README_2.md"

        else:
            filename = "README.md"

        self._create_file(filename)

        return filename

    def _create_file(self, filename: str) -> None:
        with open(filename, "w") as f:
            f.write("")

        print(f"Created markdown file: '{filename}'.")

    def _slice_contents_at_insert_position(self, method_name: str) -> tuple[str, str]:
        with open(self.filepath, "r+") as f:
            contents = f.read()

            if contents == "":
                raise ValueError(f"Unable to append to '{method_name}' - '{self.filename}' is empty")
            if contents.__contains__(method_name) is False:
                raise ValueError(f"Unable to find {method_name} in {self.filename}. Confirm '{method_name}' is found in '{self.filepath}'")

            method_start = contents.find(f">{method_name}<")
            next_method_start = contents.find("<strong", method_start)

            before = contents[0:next_method_start].strip()
            after = contents[next_method_start:-1].strip()

            return (before, after)

    def _generate_output_str(self, s1: str, s2: str) -> str:
        output = s1 + "\n\n> Example\n\n" + self.to_append + "\n" + s2
        return output



DEFAULT_ARG_TEMPLATE = """    <li>
        <b>{name} : <i>{type}, {default_arg}</i></b>
        <ul style='list-style: none'>
            <li>description</li>
        </ul>
    </li>
"""

TEMPLATE = """    <li>
        <b>{name} : <i>{type}</i></b>
        <ul style='list-style: none'>
            <li>description</li>
        </ul>
    </li>
"""



from inspect import getmembers, signature, isfunction, ismethod

from eazydocs.generator.parameters import Parameters


class Generator:
    def __init__(self, cls: object, skip_private: bool) -> None:
        self.skip_private = skip_private
        self.name = cls.__name__

        self.generate_docs(cls)

    def generate_docs(self, cls: object) -> None:
        docs = str()

        for name, member in getmembers(cls):
            if ismethod(member) or isfunction(member):
                if self._check_member(member):
                    params = Parameters(member, self.skip_private).params
                    docs += f"\n{params}\n"

        self.docs = docs
        self._fmt_class_name()

    def __repr__(self) -> str:
        return self.docs

    def _check_member(self, method: object) -> bool:
        method_name = method.__name__
        if self.skip_private:
            if method_name[0] == "_" or method_name[0:2] == "__" and method_name != "__init__":
                return False

        return self._check_params(method)

    def _check_params(self, method) -> bool:
        params = signature(method).parameters
        params = [param for param in params.keys() if param != "self"]
        if params == []:
            return False

        return True

    def _fmt_class_name(self) -> None:
        if self.docs.__contains__("__init__"):
            docs = self.docs.replace("--init--", self.name.lower())
            docs = docs.replace("__init__", self.name)
            self.docs = docs



from pandas import DataFrame


class Example:
    def __init__(self, code: str, data: DataFrame, df_shape: list[int]) -> None:
        if len(df_shape) != 2:
            raise ValueError("Length of list provided for 'df_shape' must be exactly 2 integers: [num_rows, num_columns]")

        self.data = data
        self.code = code

        self.rows = df_shape[0]
        self.columns = df_shape[1]

        self.get_example()

    def get_example(self) -> None:
        bash = "```"
        df = self._format_df()

        self.output = f"{bash}\n>>> {self.code}\n{df}\n{bash}"

    def __repr__(self) -> str:
        return self.output

    def _format_df(self) -> str:
        df = self.data

        if self.rows != -1 and self.data.shape[0] > self.rows:
            df = df[0 : self.rows]

        if self.columns != -1 and self.data.shape[1] > self.columns:
            df = df.iloc[0 : self.rows, 0 : self.columns]

        return df.to_string()



class Method:
    def __init__(self, method_name: str) -> None:
        self.fn = self._generate_function(method_name)

    def _generate_function(self, method: str) -> str:
        method = method.strip()
        method_id = method.replace("_", "-")
        return f"<strong id='{method_id}'>{method}</strong>("

    def _fmt_function(self) -> str:
        fn = self.fn.removesuffix(", ") + ")"
        return fn


from dataclasses import dataclass
from inspect import signature

from eazydocs.generator.method import Method
from eazydocs.generator.templates import DEFAULT_ARG_TEMPLATE, TEMPLATE


@dataclass
class Param:
    name: str
    arg_type: str
    default_arg: str


class Parameters(Method):
    def __init__(self, method: object, skip_private: bool) -> None:
        super().__init__(method.__name__)
        self.skip_private = skip_private

        self.generate(method)

    def generate(self, method: object) -> None:
        params = self.get_params_output(method)
        fn = self._fmt_function()

        self.params = f"{fn}\n{params}"

    def __repr__(self) -> str:
        return self.params

    def get_params_output(self, method: object) -> str:
        params = self._get_params(method)

        if params == []:
            return ""

        blockquote = "> Parameters"
        output = f"\n{blockquote}\n\n<ul style='list-style: none'>\n"

        for param in params:
            output += self._fmt_param(param)

        output += "</ul>"

        return output

    def _get_params(self, method: object) -> list[Param]:
        sig_params = signature(method).parameters

        if self.skip_private:
            parameters_filtered = [
                param for param in sig_params.keys() if param != "self" and param[0] != "_" and param != "__init__" and param[0:2] != "__"
            ]
        else:
            parameters_filtered = [param for param in sig_params.keys() if param != "self"]

        parameters = [sig_params.get(param) for param in parameters_filtered]

        params = list()

        for param in parameters:
            param = str(param).split(":")

            param_name = param[0].strip()

            # TODO: Find a way to handle parameters that are not typed but have a default value
            if len(param) < 2:
                param_arg_type = "type"
            else:
                param_arg_type = param[1].strip()

            if "=" in param_arg_type:
                param_arg_type_split = param_arg_type.split("=")
                param_arg_type = param_arg_type_split[0].strip()
                param_default_arg = param_arg_type_split[-1].strip()
            else:
                param_default_arg = ""

            params.append(Param(param_name, param_arg_type, param_default_arg))

        return params

    def _fmt_param(self, param: Param) -> None:
        if param.default_arg == "":
            template = TEMPLATE
        else:
            template = DEFAULT_ARG_TEMPLATE

        if "None" in param.default_arg:
            template = template.replace("{name}", param.name).replace("{type}", param.arg_type).replace("{default_arg}", f"optional")
            param.default_arg = "_NoDefault.no_default"
        else:
            template = (
                template.replace("{name}", param.name).replace("{type}", param.arg_type).replace("{default_arg}", f"default {param.default_arg}")
            )

        self._append_param(param)

        return template

    def _append_param(self, param: Param) -> None:
        name = param.name
        default_arg = param.default_arg

        to_append = f"<b>{name}</b>"

        if default_arg != "":
            to_append += f"<i>={default_arg}</i>, "
        else:
            to_append += ", "

        self.fn += to_append
